# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-24 15:40
Desc:
'''
from pynovice.rule_engine import RuleParser

def rule_pattern(_iterm):
    age = _iterm.get('profile_age',0)
    sms_len = _iterm.get('sms_len',0)
    micro_loan_len = _iterm.get('micro_loan_len', 0)
    endBillsLen = _iterm.get('endBillsLen', 0)
    maxDueday = _iterm.get('maxDueday', 0)
    this_apply_dueday_predicted_score = _iterm.get('this_apply_dueday_predicted_score', 0)

    _pattern = [
        'OR#old#',
            ['between#rule_10',age,18,55],
            ['>#rule_11',sms_len,50],
            ['<#rule_12',micro_loan_len,8],
            ['<', micro_loan_len, -99],
            ['and#rule_40',
                ['>',endBillsLen,4],
                ['>',maxDueday,20],
                ['>',this_apply_dueday_predicted_score,20]
             ]
    ]
    return _pattern


if __name__ == '__main__':
    iterm = {'age': 30.00, 'sms_len': 70, 'micro_loan_len': 10, 'endBillsLen': 5, 'maxDueday': 20,
             'this_apply_dueday_predicted_score': 10}

    iterm = {'userId': '154306317635289088',
               'history_loan_status': True,
               'salary': 'above 25000',
               'loan_num_1day': 1,
               'hit_device_bl': True,
               'hit_mac_bl': False,
               'hit_mobile_bl': False,
               'hit_white_list': False,
               'age': 0.0,
               'nationalId_match_name': False,
               'nationalId_used': False,
               'hit_credit_bl': False,
               'contacts_len': 3037,
               'sms_len': 1875,
               'contain_mpesa': True,
               'contain_username': True,
               'mobile_in_applicantDB': False,
               'mobile_match_set_status': True,
               'mobile_match_set_len': 0,
               'referee_in_applicantDB': False,
               'referee_match_set_status': False,
               'referee_in_applicantDB_referee': True,
               'referee_match_referee_set_status': True,
               'referee_match_referee_set_len': 3,
               'online_duration': 400,
               'company': 'Metropol',
               'gender': 'M',
               'mobiLoanScore': 0,
               'is_unemployed': False,
               'micro_loan_len': 0,
               'user_category': 'old',
               'apply_hour': 0,
               'frequent_contact': 0,
               'simi_score': 0.0,
               'faceNum': -1,
               'faceLiveness': 0.0,
               'max_levenshtein': 0,
               'opesa_maxdueday': -1,
               'opesa_waitdueday': -1,
               'total_borrowing': 2500.0,
               'is_smsmsg': 0,
               'interaction_order': 1,
               'score': 0,
               'pass': True,
               'reason': 'ok',
               'apply_period_length': 91,
               'tags': '[]',
               'gaid_cnt': 2,
               'email_duplicate_cnt': 2,
               'referee_mobile_duplicate_cnt': 3,
               'today_passed_risk_info': 0,
               'latest_balance_in_sms': 0.84,
               'idTimes': 0,
               'maritalStatus': 'Single',
               'hasOutstandingLoan': 'No',
               'employmentDetails': 'Permanent',
               'residentialStatus': 'Rented',
               'eduLevel': 'Graduate',
               'mobile_no': '254703818484',
               'isBtc': 0,
               'isAdmin': 0,
               'num_is_admin_apply': 0,
               'num_is_admin_passed': 0,
               'num_is_btc_apply': 0,
               'num_is_btc_passed': 0,
               'gps_cell_token': '182f476c',
               'apply_term': 2,
               'hit_grey_mobile': False,
               'zh_sms_len': 0,
               'apply_amount': 2500.0,
               'profile_age': 26.063013698630137,
               'gps_tile_new_apply_change_rate': -0.9999850002249967,
               'gps_tile_new_apply_passrate_change_rate': 0.0,
               'gps_tile_today_new_apply': 0.0,
               'gps_tile_today_new_passrate': 0.0,
               'has_android_root_tools': None,
               'app_contains_whatsapp': True,
               'sms_sent_cnt': 778,
               'sms_sent_ratio': 0.41,
               'hit_sensitive_content': True,
               'hit_multiple_loans': 0.0,
               'hit_multiple_loans_d15': 0.0,
               'hit_student_content': True,
               'sms_contains_username': False,
               'affiliate_waitduebill_cnt': 0,
               'affiliate_waitbill_cnt': 0,
               'loan_history_service_is_live': True,
               'affiliate_maxdueday': -1,
               'affiliate_endbills_cnt_max': 1,
               'hit_sensitive_title': False,
               'num_firstloan_passed': 0,
               'num_reloan_passed': 0,
               'num_longperiod_passed': 0,
               'num_reloan_longperiod_passed': 0,
               'num_rule_19_passed': 0,
               'num_rule_38_passed': 0,
               'num_ok_19_passed': 0,
               'num_ok_38_passed': 0,
               'num_ok_passed': 0,
               'num_old_passed': 0,
               'num_old_19_passed': 0,
               'hit_internal_bl': True,
               'hit_internal_bl_by_manual': False,
               'black_nationalid_service_is_live': True,
               'emergency_is_self': False,
               'emergency_contact_in_contacts': True,
               'gps_is_empty': False,
               'gps_cheat': False,
               'gps_covid19': False,
               'hours_passed_since_apply': 1.985,
               'hit_loaned_content': 0,
               'bank_transfer_in_times': 1,
               'overpaid_amount': 0,
               'kenya': {'num_firstloan_passed': 2, 'num_reloan_passed': 0},
               '_risk_scene': 'ReLoanRiskEvent',
               'risk_passed': True,
               'maxDueday': 50,
               'endBillsLen': 2,
               'is_manual': False,
               'this_apply_dueday_predicted': -999,
               'this_apply_dueday_predicted_score': 0.0,
               'cache': {'okash': {'stats:pass': {'num_firstloan_passed': 0,
                                                  'num_reloan_passed': 0,
                                                  'num_longperiod_passed': 0,
                                                  'num_reloan_longperiod_passed': 0,
                                                  'num_rule_19_passed': 0,
                                                  'num_rule_38_passed': 0,
                                                  'num_ok_19_passed': 0,
                                                  'num_ok_38_passed': 0,
                                                  'num_ok_passed': 0,
                                                  'num_old_passed': 0,
                                                  'num_old_19_passed': 0,
                                                  'app': 'OKash'}},
                         'opesa': {'loan:history': {'maxdueday': -1,
                                                    'waitduebill_cnt': 0,
                                                    'waitbill_cnt': 0,
                                                    'waitmaxdueday': -999,
                                                    'waitbilltotal': '0.0',
                                                    'endbill_cnt': 1,
                                                    'app': 'OPesa'},
                                   'stats:pass': {'num_firstloan_passed': 0,
                                                  'num_reloan_passed': 0,
                                                  'num_longperiod_passed': 0,
                                                  'num_reloan_longperiod_passed': 0,
                                                  'num_rule_19_passed': 0,
                                                  'num_rule_38_passed': 0,
                                                  'num_ok_19_passed': 0,
                                                  'num_ok_38_passed': 0,
                                                  'num_ok_passed': 0,
                                                  'num_old_passed': 0,
                                                  'num_old_19_passed': 0,
                                                  'app': 'OPesa'}},
                         'credithela': {'loan:history': {'maxdueday': -999,
                                                         'waitduebill_cnt': 0,
                                                         'waitbill_cnt': 0,
                                                         'waitmaxdueday': -999,
                                                         'waitbilltotal': -999,
                                                         'endbill_cnt': 0,
                                                         'app': 'CreditHela'},
                                        'stats:pass': {'num_firstloan_passed': 2,
                                                       'num_reloan_passed': 0,
                                                       'num_longperiod_passed': 0,
                                                       'num_reloan_longperiod_passed': 0,
                                                       'num_rule_19_passed': 0,
                                                       'num_rule_38_passed': 0,
                                                       'num_ok_19_passed': 0,
                                                       'num_ok_38_passed': 0,
                                                       'num_ok_passed': 2,
                                                       'num_old_passed': 0,
                                                       'num_old_19_passed': 0,
                                                       'app': 'CreditHela'}}}}

    rule = rule_pattern(iterm)

    engine = RuleParser()
    out = engine.evaluate(_rule=rule)
    print(engine.get_hit_rules())
    print(out)


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-28 11:06
Desc:
    1.类别特征转换为数值，[进行分箱后]，2.再次将分箱转换为类别字典
'''
import pandas as pd

class CateCoder:
    def __init__(self):
        self.cate_dict={}

    def fit(self,df_field, df_label,target_label=1):
        '''
            类别字段按照坏样本占比排序编码
            :param df_field:
            :param df_label:
            :param target_label:
            :return:
            cate_dict: {field0:0,field1:1,field2:2}
            '''
        df_field.reset_index(drop=True, inplace=True)
        df_label.reset_index(drop=True, inplace=True)
        df_concat = pd.concat([df_field, df_label], axis=1)
        df_concat.columns = ['field', 'label']
        bad_rate_sorted = df_concat.groupby('field').agg(lambda x: sum(x == target_label) / len(x)).sort_values('label')

        rever_cate_dict = dict(enumerate(bad_rate_sorted.index))  # {0:field0,1:field1,2:field2}
        self.cate_dict = {v: k for k, v in rever_cate_dict.items()}  # {field0:0,field1:1,field2:2}

        # df_field = df_field.apply(lambda x: self.cate_dict.get(x, -1))
        # return df_field

    def transform(self,x):
        return self.cate_dict.get(x, -1)

    def fit_transform(self,df_field, df_label, target_label=1):
        self.fit(df_field, df_label, target_label)
        df_out = df_field.apply(lambda x: self.transform(x))
        return df_out

    # tmp
    def blocks_2_catedict(self,blocks):
        # 转换原始类别标签到分箱编码
        cate_code = list(self.cate_dict.values())
        block_code = pd.cut(cate_code, bins=blocks)#, labels=list(range(len(blocks) - 1)))
        block_dict = dict(zip(cate_code, block_code))
        return block_dict

if __name__ == '__main__':
    from pynovice.score_card.src.data_boxing import chi_blocks
    df = pd.DataFrame([[1, 2, 3, 4, 5, 5, 5, 3, 3, 3, 2, 1, 5, 7], [1, 1, 0, 0, 0, 0, 0, 1, 1, 0, 1, 1, 1, 1]]).T
    df.columns = ['field', 'label']
    df_field, df_label = df['field'], df['label']

    cate_coder = CateCoder()
    df_field = cate_coder.fit_transform(df_field, df_label,target_label=1)
    x = cate_coder.transform(5)
    print(df_field,x)


    block = chi_blocks(df_field, df_label, box_num=5, dfree=4, cf=0.1)
    print(block)
    block_dict=cate_coder.blocks_2_catedict(block)
    print(block_dict)






# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-07-03 17:31
Desc:
'''

# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-16 18:09
Desc:
map extends
'''
import webbrowser
import os
import pandas as pd
from pyecharts.components import Table
from pyecharts import options as opts
from pyecharts.charts import Bar, Line, Tab, Page

CURRENT_PATH = os.path.dirname(os.path.abspath(__file__))
class DistributionAnysis:
    def __init__(self, image_separation=False):
        self.commit = "Analyze and plot the data "
        self.series_name_add = '_' if image_separation else ''
        self.html_name = os.path.abspath('data_distribution.html')
        self.datas = []

    def get_plot_base(self, _df, title_name='数据分布', subtitle_name='',chart_type=None,xname=None,yname=None):
        '''
        plot line & bar info
        :param _df:
        :return:
        '''
        bar = Bar()
        line = Line()
        bar.set_global_opts(title_opts=opts.TitleOpts(title=title_name, subtitle=subtitle_name),
                            yaxis_opts=opts.AxisOpts(axislabel_opts=opts.LabelOpts(formatter="{value}"),name=yname),
                            datazoom_opts=opts.DataZoomOpts(),  # 滑动条
                            xaxis_opts=opts.AxisOpts(axislabel_opts=opts.LabelOpts(rotate=-15),name=xname),  # label旋转
                            toolbox_opts=opts.ToolboxOpts(),
                            )
        xaxis = list(map(str, _df.index))
        bar.add_xaxis(xaxis)
        line.add_xaxis(xaxis)
        for field in _df.columns:
            field=str(field)
            if chart_type=='line':
                line.add_yaxis(series_name=field, y_axis=_df[field].to_list(),is_symbol_show=False)
            elif chart_type=='bar':
                bar.add_yaxis(series_name=field, yaxis_data=_df[field].to_list(), is_selected=True,)
            else:
                bar.add_yaxis(series_name=field, yaxis_data=_df[field].to_list(), is_selected=True)
                line.add_yaxis(series_name=field+self.series_name_add, y_axis=_df[field].to_list(), is_symbol_show=True)
        bar.overlap(line)
        return bar

    def get_table_base(self, _df, title_name='数据分布', subtitle_name=''):
        '''
        数据详情表 （未使用）
        :param _df:
        :param title_name:
        :param subtitle_name:
        :return:
        '''
        _df = _df.reset_index()
        table = Table()
        headers = _df.columns.tolist()
        rows = _df.values
        table.add(headers, rows).set_global_opts(
            title_opts=opts.ComponentTitleOpts(title="Table-{}".format(title_name), subtitle=subtitle_name)
        )
        return table

    def add_data(self,_df,title_name, subtitle_name='',chart_type=None,xname=None,yname=None):
        _plot = self.get_plot_base(_df, title_name, subtitle_name,chart_type,xname,yname)
        self.datas.append(_plot)

    def plot_summary(self,page_name='',page_type='page',web_open=False,html_name=None):
        assert page_type in ('page','tab'),'仅支持单页多图(page)和选项卡多图(tab)'
        if html_name:
            self.html_name = os.path.abspath(str(html_name)+'.html')
        if 'page' == page_type:
            page = Page(page_title=page_name)  # 单页多图
            for idata in self.datas:
                page.add(idata)
                page.render(self.html_name)
        else:
            tab = Tab(page_title=page_name, )  # 选项卡多图
            for idata in self.datas:
                tab_name = idata.options.get('title').opts[0].get('text')
                tab.add(idata,tab_name)
                tab.render(self.html_name)
        if web_open:
            webbrowser.open(url='file://' + self.html_name, new=0, autoraise=True)

if __name__ == '__main__':
    df_raw = pd.DataFrame( data = [[-0.4302744, 35.5331974],
                                   [-1.21167  , 36.8973031],
                                   [ 0.5217472, 35.255192 ],
                                   [-2.2326433, 38.0769783],
                                   [-1.0625246, 34.4734754]],
                           columns=['latitude', 'longitude'])

    any = DistributionAnysis(image_separation=True)
    any.add_data(_df=df_raw,title_name='aa1', subtitle_name='aa',chart_type='bar',yname='xname',xname='yname')
    any.add_data(_df=df_raw,title_name='bb1', subtitle_name='bb')
    any.plot_summary(page_name='1234',page_type='tab',web_open=True,html_name=None)


import pandas as pd
import re
import time
import datetime
import pickle
from dateutil.relativedelta import relativedelta
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
import seaborn as sns
from statsmodels.stats.outliers_influence import variance_inflation_factor
from sklearn.linear_model import LogisticRegressionCV
import statsmodels.api as sm
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import roc_auc_score
import numpy as np

### 计算KS值
def KSs(df, score, target):
    '''
    :param df: 包含目标变量与预测值的数据集
    :param score: 得分或者概率
    :param target: 目标变量
    :return: KS值
    '''
    total = df.groupby([score])[target].count()
    bad = df.groupby([score])[target].sum()
    all = pd.DataFrame({'total':total, 'bad':bad})
    all['good'] = all['total'] - all['bad']
    all[score] = all.index
    all = all.sort_values(by=score,ascending=False)
    all.index = range(len(all))
    all['badCumRate'] = all['bad'].cumsum() / all['bad'].sum()
    all['goodCumRate'] = all['good'].cumsum() / all['good'].sum()
    KS = all.apply(lambda x: x.badCumRate - x.goodCumRate, axis=1)
    return max(KS)

#读取数据
trainData = pd.read_csv('test1.csv')
##只计算统计变量
allFeatures = list(trainData.columns)
allFeatures.remove('Unnamed: 0')

## 上一步的结果
multi_analysis = []

### (1)将多变量分析的后变量带入LR模型中
y = trainData['y']
X = trainData[multi_analysis]
X['intercept'] = [1] * X.shape[0]

LR = sm.Logit(y, X).fit()
summary = LR.summary()
print(summary)
pvals = LR.pvalues
pvals = pvals.to_dict()
print(pvals)

# 有些变量不显著，需要逐步剔除
varLargeP = {k: v for k, v in pvals.items() if v >= 0.1}
varLargeP = sorted(varLargeP.items(), key=lambda d: d[1], reverse=True)
while (len(varLargeP) > 0 and len(multi_analysis) > 0):
    # 每次迭代中，剔除最不显著的变量，直到
    # (1) 剩余所有变量均显著
    # (2) 没有特征可选
    varMaxP = varLargeP[0][0]
    print(varMaxP)
    if varMaxP == 'intercept':
        print('the intercept is not significant!')
        break
    multi_analysis.remove(varMaxP)
    y = trainData['result']
    X = trainData[multi_analysis]
    X['intercept'] = [1] * X.shape[0]

    LR = sm.Logit(y, X).fit()
    pvals = LR.pvalues
    pvals = pvals.to_dict()
    varLargeP = {k: v for k, v in pvals.items() if v >= 0.1}
    varLargeP = sorted(varLargeP.items(), key=lambda d: d[1], reverse=True)

summary = LR.summary()
trainData['prob'] = LR.predict(X)
ks = KSs(trainData, 'prob', 'result')
auc = roc_auc_score(trainData['result'], trainData['prob'])
print('normalLR:ks {}, auc {}'.format(ks, auc)) 

# 将模型保存
saveModel = open('LR_Model_Normal.pkl', 'wb+')
pickle.dump(LR, saveModel)
saveModel.close()


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-23 17:59
Desc:
    卡方分箱 【迭代依次查找（合并分箱后卡方最小的两个箱)进行合并】
    连续特征值观测值过多的可以先进行等频率分箱
    from https://zhuanlan.zhihu.com/p/115267395
    1.初始化阶段：首先按照属性值的大小进行排序（对于非连续特征，需要先做数值转换，比如转为坏人率，然后排序），然后每个属性值单独作为一组。

    2、合并阶段：
    （1）对每一对相邻的组，计算卡方值。
    （2）根据计算的卡方值，对其中最小的一对邻组合并为一组。
    （3）不断重复（1），（2）直到计算出的卡方值都不低于事先设定的阈值，或者分组数达到一定的条件（如最小分组数5，最大分组数8）并且每箱都包含正负多个label。

    >>> df = pd.DataFrame([[1,2,3,4,5,5,5,3,3,3,2,1,5,7],[1,1,0,0,0,0,0,1,1,0,1,1,1,1]]).T
    >>> df.columns=['field','label']
    >>> df_field, df_label = df['field'],df['label']
    >>> aa = chi_blocks(df_field, df_label, box_num=5, dfree=4, cf=0.1)
    >>> print(aa)
    [-inf, 4.0, 5.0, inf]

'''

import pandas as pd
import numpy as np
from scipy.stats import chi2
from pynovice.score_card.src.data_boxing import frequence_blocks

def get_chi2(df_field,df_label):
    '''
    计算卡方统计量
    :param df_field: 需分箱的字段 Series
    :param df_label: 样本的标签 Series
    :return:
        各箱样本频率,期望频率,卡发值
    '''

    df_field.reset_index(drop=True,inplace=True)
    df_label.reset_index(drop=True,inplace=True)
    df_concat = pd.concat([df_field,df_label],axis=1)
    df_concat.columns = ['field','label']
    df_concat.sort_values(by='field',ascending= True,inplace=True)

    dict_label_ratio = dict(df_label.value_counts() / df_label.count())
    _group = df_concat.groupby('field')

    group_cnt = []
    group_except_cnt = []
    for i,j in dict_label_ratio.items():
        iterm = _group.agg([(i, lambda x: sum(x == i))])
        group_cnt.append(iterm)
        iterm_expect = _group.agg([(i, lambda x: len(x)*j)])
        group_except_cnt.append(iterm_expect)
    df_pos_cnt = pd.concat(group_cnt,axis=1).droplevel(level=0,axis=1)  # 样本频率
    df_except_cnt = pd.concat(group_except_cnt,axis=1).droplevel(level=0,axis=1)  # 样本期望频率
    df_chi_square = (df_pos_cnt - df_except_cnt) ** 2 / df_except_cnt  # 卡方值

    return df_pos_cnt,df_except_cnt,df_chi_square


def cal_chisqure_threshold(dfree=4, cf=0.1):
    '''
    根据给定的自由度和显著性水平, 计算卡方阈值
    '''
    percents = [0.95, 0.90, 0.5, 0.1, 0.05, 0.025, 0.01, 0.005]

    ## 计算每个自由度，在每个显著性水平下的卡方阈值
    df = pd.DataFrame(np.array([chi2.isf(percents, df=i) for i in range(1, 30)]))
    df.columns = percents
    df.index = df.index + 1

    pd.set_option('precision', 3)
    return df.loc[dfree, cf]

def chi_merge(_pos_cnt, _except_cnt, window=2):
    '''
    合并相邻两bins，并计算合并后的卡方值，得到合并后最小卡方值和分箱位置索引
    :param _pos_cnt:
    :param _except_cnt:
    :param window:
    :return:
        合并分箱后的数据及对应的最小卡方值
    '''
    df_pos_cnt = _pos_cnt.sort_index().copy()
    df_except_cnt = _except_cnt.sort_index().copy()

    df_bins_init = df_pos_cnt.rolling(window=window).sum()[1:]
    df_expect_bins_init = df_except_cnt.rolling(window=window).sum()[1:]
    chi_bins = ((df_bins_init-df_expect_bins_init)**2/df_expect_bins_init).sum(axis=1).reset_index(drop=True)
    min_chi = chi_bins.min()
    idx_chi = chi_bins.idxmin()

    # merge
    index = df_pos_cnt.index
    df_pos_cnt.iloc[idx_chi+1,:] = df_pos_cnt.iloc[idx_chi,:]+df_pos_cnt.iloc[idx_chi+1,:]
    df_pos_cnt.drop(index=index[idx_chi],inplace=True)

    df_except_cnt.iloc[idx_chi+1,:] = df_except_cnt.iloc[idx_chi,:]+df_except_cnt.iloc[idx_chi+1,:]
    df_except_cnt.drop(index=index[idx_chi],inplace=True)

    return  df_pos_cnt,df_except_cnt,min_chi


def oneclass_merge(df_pos_cnt, df_except_cnt,sample_threshold=2):
    '''
    检查是否有箱没有好或者坏样本或者只有一类的。如果有，需要跟相邻的箱进行合并，直到每箱同时包含多类别样本
    :param df_pos_cnt:
    :param df_except_cnt:
    :param sample_threshold:bins中至少含有样本阈值
    :return:

    '''
    raw_index = df_pos_cnt.index
    df_pos_cnt.reset_index(drop=True,inplace=True)
    df_except_cnt.reset_index(drop=True,inplace=True)
    df_oneclass = df_pos_cnt[df_pos_cnt.apply(lambda x: (1==sum(x > 0)) or (sum(x)<sample_threshold), axis=1)]
    idx_oneclass = df_oneclass.index.to_list()

    if len(idx_oneclass)>0&(len(df_pos_cnt)>1):
        icur = idx_oneclass[0]
        if icur == 0:  # 如果分箱区间在最前,则向下合并
            df_pos_cnt.loc[icur + 1, :] = df_pos_cnt.loc[icur + 1, :] + \
                                          df_pos_cnt.loc[icur, :]
            df_except_cnt.loc[icur + 1, :] = df_except_cnt.loc[icur + 1, :] + \
                                             df_except_cnt.loc[icur, :]
            drop_index = icur
        elif icur == len(raw_index) - 1:  # 如果分箱区间在最后，则向上合并
            df_pos_cnt.loc[icur - 1, :] = df_pos_cnt.loc[icur - 1, :] + \
                                          df_pos_cnt.loc[icur, :]
            df_except_cnt.loc[icur - 1, :] = df_except_cnt.loc[icur - 1, :] + \
                                             df_except_cnt.loc[icur, :]
            drop_index = icur
        else:  # 如果分箱区间在中间，则判断与其相邻的最小卡方的区间，然后进行合并
            select_idx = [icur - 1, icur, icur + 1]
            pos_cnt, except_cnt, _ = chi_merge(df_pos_cnt.loc[select_idx, :], df_except_cnt.loc[select_idx, :], window=2)
            df_pos_cnt.loc[pos_cnt.index, :] = pos_cnt
            df_except_cnt.loc[pos_cnt.index, :] = except_cnt
            if max(pos_cnt.index) < icur + 1:
                drop_index = icur + 1
            elif min(pos_cnt.index) > icur - 1:
                drop_index = icur - 1
            else:
                drop_index = icur
        df_pos_cnt.drop(index=drop_index, inplace=True)
        df_except_cnt.drop(index=drop_index, inplace=True)
    df_pos_cnt.index = raw_index[df_pos_cnt.index]
    df_except_cnt.index = raw_index[df_except_cnt.index]
    if len(idx_oneclass) > 1:
        return oneclass_merge(df_pos_cnt, df_except_cnt)
    return df_pos_cnt,df_except_cnt

def chi_blocks(df_field, df_label, box_num=5, dfree=4, cf=0.1,max_iterations=100):
    '''
    卡方分箱
    :param df_field: 需分箱的字段 Series
    :param df_label: 样本的标签 Series
    :param dfree: 卡方分布的自由度
    :param cf: 卡方分布的显著性水平
    :param box_num: 最大分箱数
    :param sample_threshold:  bins中最少样本数
    :param _type:  number|category
    :return:
        分箱的eges
    '''
    if box_num<2:
        return [-np.inf,np.inf]
    chi_threshold = cal_chisqure_threshold(dfree, cf)  # 卡方阈值
    sample_threshold = df_field.shape[0] / 20  # 总体的5%(bins要大于5%)
    iterations_bins = box_num+max_iterations
    if len(set(df_field))>iterations_bins: #连续特征太多，进行初步分bins
        init_bins = frequence_blocks(x=df_field.to_list(), bins=iterations_bins)
        _bins = init_bins[1:-1]+[init_bins[-2]+10]
        df_field = pd.cut(df_field,bins=init_bins,labels=_bins)

    pos_cnt, except_cnt,_ = get_chi2(df_field, df_label)  # 各箱样本频率和期望频率
    df_pos_cnt, df_except_cnt = oneclass_merge(pos_cnt, except_cnt,sample_threshold)
    min_chi = -np.inf
    # 如果变量区间的最小卡方值小于阈值，则继续合并直到最小值大于等于阈值
    while (min_chi<chi_threshold and len(df_pos_cnt) >= box_num):
        _pos_cnt, _except_cnt, min_chi = chi_merge(df_pos_cnt, df_except_cnt, window=2)
        df_pos_cnt, df_except_cnt = oneclass_merge(_pos_cnt, _except_cnt,sample_threshold)

    blocks = np.concatenate([[-np.inf],df_pos_cnt.index,[np.inf]])
    return blocks.tolist()

if __name__ == '__main__':
    df = pd.DataFrame([[1,2,3,4,5,5,5,3,3,3,2,1,5,7],[1,1,0,0,0,0,0,1,1,0,1,1,1,1]]).T
    df.columns=['field','label']
    df_field, df_label = df['field'],df['label']
    aa = chi_blocks(df_field, df_label, box_num=5, dfree=4, cf=0.1)
    print(aa)



# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-03-25 14:40
Desc:
'''

from bert4keras.models import build_transformer_model
from bert4keras.tokenizers import Tokenizer
import numpy as np
import os
import requests
import zipfile
from pynovice.util import progress_bar

CUR_PASTH = os.path.dirname(os.path.abspath(__file__))

class ExtractBertFeatures:
    def __init__(self,model_name = 'uncased_L-4_H-256_A-4'):
        bert_model_dir = os.path.join(CUR_PASTH,'data',model_name)
        self.load_model(model_dir=bert_model_dir)
        pass

    def load_model(self,model_dir):
        config_name = os.path.join(model_dir, 'bert_config.json')
        checkpoint_name = os.path.join(model_dir, 'bert_model.ckpt')
        vocab_name = os.path.join(model_dir, 'vocab.txt')
        self.tokenizer = Tokenizer(vocab_name, do_lower_case=True)  # 建立分词器
        self.model = build_transformer_model(config_name, checkpoint_name)  # 建立模型，加载权重

    def predict(self,x,second_text=None,max_length=None,first_length=None,second_length=None,use_multiprocessing=False):
        token_ids, segment_ids = self.tokenizer.encode(x,second_text,max_length,first_length,second_length)
        features = self.model.predict([np.array([token_ids]), np.array([segment_ids])])
        return features

class DownloadBertModel:
    def __init__(self,url='https://storage.googleapis.com/bert_models/2020_02_20/uncased_L-4_H-256_A-4.zip'):
        self.model_url = url
        self.name = url.split('/')[-1].split('.')[0]
        self.save_parh = os.path.join(CUR_PASTH, 'data', self.name)

    def dump(self, save_parh=''):
        if not save_parh:
            save_parh = self.save_parh
        if not os.path.exists(save_parh):
            os.makedirs(save_parh)
        # 模型下载
        url = self.model_url
        with requests.get(url, stream=True) as r:
            filesize = r.headers["Content-Length"]
            chunk_size = 1024
            times = int(filesize) // chunk_size
            print("模型文件大小:", filesize, "bytes")
            print("下载地址:", url)
            print('-' * 30)
            print("开始下载")
            with open("_tmp.zip", "wb") as fw:
                start = 0
                for chunk in r.iter_content(chunk_size):
                    fw.write(chunk)
                    start += 1
                    progress_bar(start, times)
        # 解压文件
        f = zipfile.ZipFile('_tmp.zip', 'r')
        for file in f.namelist():
            f.extract(file,save_parh)
        os.remove("_tmp.zip")

if __name__ == '__main__':
    # aa=DownloadBertModel()
    # aa.dump()
    print(CUR_PASTH)
    bb=ExtractBertFeatures()
    print(bb.predict('你好啊'))



import pandas as pd
import numpy as np
import pickle
from sklearn.ensemble import GradientBoostingClassifier
from sklearn import cross_validation, metrics
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model.logistic import LogisticRegression
import warnings
warnings.filterwarnings("ignore")
'''
之前所需要进行的分箱或者预处理操作暂略
'''

def KS(df, score, target):
    '''
    :param df: 包含目标变量与预测值的数据集
    :param score: 得分或者概率
    :param target: 目标变量
    :return: KS值
    '''
    total = df.groupby([score])[target].count()
    bad = df.groupby([score])[target].sum()
    all = pd.DataFrame({'total':total, 'bad':bad})
    all['good'] = all['total'] - all['bad']
    all[score] = all.index
    all = all.sort_values(by=score,ascending=False)
    all.index = range(len(all))
    all['badCumRate'] = all['bad'].cumsum() / all['bad'].sum()
    all['goodCumRate'] = all['good'].cumsum() / all['good'].sum()
    KS = all.apply(lambda x: x.badCumRate - x.goodCumRate, axis=1)
    return max(KS)

#读取数据
trainData = pd.read_csv('test1.csv')
##只计算统计变量
allFeatures = list(trainData.columns)
allFeatures.remove('Unnamed: 0')
allFeatures.remove('result')


trainData, testData = train_test_split(trainData,test_size=0.4,random_state=0)

X = np.matrix(trainData[allFeatures])
y = trainData['result']


# 未经调参进行GBDT模型训练
gbm0 = GradientBoostingClassifier(random_state=10)
gbm0.fit(X,y)

y_pred = gbm0.predict(X)
y_predprob = gbm0.predict_proba(X)[:,1].T
trainData['probb'] = y_predprob
print ("Accuracy : %.4g" % metrics.accuracy_score(y, y_pred)) # 0.8919
print ("AUC Score (Train): %f" % metrics.roc_auc_score(np.array(y.T), y_predprob)) # 0.
print ("KS is :{}".format(KS(trainData, 'probb', 'gbflag')))

X_test = np.matrix(testData[allFeatures])
y_test = np.matrix(testData['result']).T
#在测试集上测试GBDT性能
y_predtest = gbm0.predict(X_test)
y_predprobtest = gbm0.predict_proba(X_test)[:,1].T
testData['predprob'] = list(y_predprobtest)
print ("Accuracy : %.4g" % metrics.accuracy_score(y_test, y_predtest))
print ("AUC Score (Test): %f" % metrics.roc_auc_score(np.array(y_test)[:,0], y_predprobtest))
print ("KS is :{}".format(KS(testData, 'predprob', 'gbflag')))



'''
GBDT调参
'''
# 1, 选择较小的步长(learning rate)后，对迭代次数(n_estimators)进行调参

X = pd.DataFrame(X)

param_test1 = {'n_estimators':range(20,81,10)}
gsearch1 = GridSearchCV(estimator = GradientBoostingClassifier(learning_rate=0.1, min_samples_split=30,
                                  min_samples_leaf=5,max_depth=8,max_features='sqrt', subsample=0.8,random_state=10),
                       param_grid = param_test1, scoring='roc_auc',iid=False,cv=5)
gsearch1.fit(X,y)
best_n_estimator = gsearch1.best_params_['n_estimators']


# 2, 对决策树最大深度max_depth和内部节点再划分所需最小样本数min_samples_split进行网格搜索
param_test2 = {'max_depth':range(3,16), 'min_samples_split':range(2,10)}
gsearch2 = GridSearchCV(estimator = GradientBoostingClassifier(learning_rate=0.1, n_estimators=best_n_estimator, min_samples_leaf=20, max_features='sqrt', subsample=0.8, random_state=10),
                        param_grid = param_test2, scoring='roc_auc',iid=False, cv=5)
gsearch2.fit(X,y)
best_max_depth = gsearch2.best_params_['max_depth']



#3, 再对内部节点再划分所需最小样本数min_samples_split和叶子节点最少样本数min_samples_leaf一起调参
param_test3 = {'min_samples_split':range(10,101,10), 'min_samples_leaf':range(5,51,5)}
gsearch3 = GridSearchCV(estimator = GradientBoostingClassifier(learning_rate=0.1, n_estimators=best_n_estimator,max_depth=best_max_depth,
                                     max_features='sqrt', subsample=0.8, random_state=10),
                       param_grid = param_test3, scoring='roc_auc',iid=False, cv=5)
gsearch3.fit(X,y)
best_min_samples_split, best_min_samples_leaf = gsearch3.best_params_['min_samples_split'],gsearch3.best_params_['min_samples_leaf']



#4, 对最大特征数max_features进行网格搜索,注意max_features 必须小于最大的树值
param_test4 = {'max_features':range(5,best_min_samples_leaf+1,5)}
gsearch4 = GridSearchCV(estimator = GradientBoostingClassifier(learning_rate=0.1, n_estimators=best_n_estimator,max_depth=best_max_depth, min_samples_leaf =best_min_samples_leaf,
               min_samples_split =best_min_samples_split, subsample=0.8, random_state=10),
                       param_grid = param_test4, scoring='roc_auc',iid=False, cv=5)
gsearch4.fit(X,y)
best_max_features = gsearch4.best_params_['max_features']

#5, 对采样比例进行网格搜索
param_test5 = {'subsample':[0.6+i*0.05 for i in range(8)]}
gsearch5 = GridSearchCV(estimator = GradientBoostingClassifier(learning_rate=0.1, n_estimators=best_n_estimator,max_depth=best_max_depth,
                                                               min_samples_leaf =best_min_samples_leaf, max_features=best_max_features,random_state=10),
                       param_grid = param_test5, scoring='roc_auc',iid=False, cv=5)
gsearch5.fit(X,y)
best_subsample = gsearch5.best_params_['subsample']

gbm_best = GradientBoostingClassifier(learning_rate=0.1, n_estimators=best_n_estimator,max_depth=best_max_depth,
                                      min_samples_leaf =best_min_samples_leaf, max_features=best_max_features,subsample =best_subsample, random_state=10)
gbm_best.fit(X,y)

#在测试集上测试并计算性能
y_pred = gbm_best.predict(X_test)
y_predprob = gbm_best.predict_proba(X_test)[:,1].T
testData['predprob'] = list(y_predprob)
print ("Accuracy : %.4g" % metrics.accuracy_score(y_test, y_pred))
print ("AUC Score (Test): %f" % metrics.roc_auc_score(np.array(y_test)[:,0], y_predprob))
print ("KS is :{}".format(KS(testData, 'predprob', 'gbflag')))


#GBDT模型生成特征

# 利用GBDT的结果衍生新的特征
new_feature= gbm_best.apply(X)[:,:,0]
grd_enc = OneHotEncoder()
grd_enc.fit(new_feature)
x = grd_enc.transform(new_feature)
classifier=LogisticRegression()
classifier.fit(x,y)

new_feature_test= gbm_best.apply(X_test)[:,:,0]
x_test = grd_enc.transform(new_feature_test)
y_pred_lr = classifier.predict_proba(x_test)[:,1]
lr_pred = pd.DataFrame({'predprob':y_pred_lr, 'gbflag': np.array(y_test)[:,0]})

print ("AUC Score (Test): %f" % metrics.roc_auc_score(np.array(y_test)[:,0], y_pred_lr))
print ("KS is :%f"%KS(lr_pred, 'predprob', 'gbflag'))

# 将模型保存
saveModel = open('./model/GBDT_statsmodels_GBDT.pkl', 'wb+')
pickle.dump(gbm_best, saveModel)
saveModel.close()




# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-24 10:55
Desc:
    简易的风控规则引擎 (仅供自己测试)
    ["操作符"，["操作符1#tag"， "参数1"， "参数2"， ...]，["操作符2"， "参数1"， "参数2"， ...]]
    >>> rules = ["and",[">#rule_10", 0, 0.05],["in#rule_20", 3, 2, 3, 4]]
    (False, [{'rule_10': False}, {'rule_20': True}, {'untag': False}])

'''
from numpy import unicode
import json

class LogicalFunctions(object):
    '''
    定义的方法集合
    '''
    def __init__(self):
        self.func_ = {
            '=': self.eq,
            '!=': self.neq,
            '>': self.gt,
            '>=': self.gte,
            '<': self.lt,
            '<=': self.lte,
            'and': self.and_,
            'in': self.in_,
            'between': self.between_,
            'or': self.or_,
            'not': self.not_,
            'str': self.str_,
            'int': self.int_,
            '+': self.plus,
            '-': self.minus,
            '*': self.multiply,
            '/': self.divide
        }

    def eq(self, *args):
        return args[0] == args[1]

    def neq(self, *args):
        return args[0] != args[1]

    def in_(self, *args):
        return args[0] in args[1:]

    def between_(self, *args):
        return (args[0] >= args[1]) and (args[0] <= args[2])

    def gt(self, *args):
        return args[0] > args[1]

    def gte(self, *args):
        return args[0] >= args[1]

    def lt(self, *args):
        return args[0] < args[1]

    def lte(self, *args):
        return args[0] <= args[1]

    def not_(self, *args):
        return not args[0]

    def or_(self, *args):
        return any(args)

    def and_(self, *args):
        return all(args)

    def int_(self, *args):
        return int(args[0])

    def str_(self, *args):
        return unicode(args[0])

    def upper(self, *args):
        return args[0].upper()

    def lower(self, *args):
        return args[0].lower()

    def plus(self, *args):
        return sum(args)

    def minus(self, *args):
        return args[0] - args[1]

    def multiply(self, *args):
        return args[0] * args[1]

    def divide(self, *args):
        return float(args[0]) / float(args[1])

    def abs(self, *args):
        return abs(args[0])

class RuleParser(object):
    def __init__(self):
        self.funcSet = LogicalFunctions().func_

    def get_func(self,_func):
        func = self.funcSet.get(_func)
        if not func:
            self.func_list()
            func = lambda _:None
        return func

    def func_list(self):
        _func_list = list(self.funcSet.keys())
        print("supported function list: {}".format(_func_list))

    def validate_check(self,rule):
        if isinstance(rule, str):
            rule = json.loads(rule)
        else:
            rule = rule
        if not isinstance(rule, list):
            raise ('Rule must be a list, got {}'.format(type(rule)))
        if len(rule) < 2:
            raise ('Must have at least one argument.')
        return rule

    def evaluate(self,_rule):
        self.whole_result = []
        rule = self.validate_check(_rule)
        result = self._evaluate(rule)
        return result,self.whole_result

    # filter hit_rule
    def get_hit_rules(self,exclude_rules=[]):
        result = self.whole_result
        filt_out = []
        for irule in result:
            for _k,_v in irule.items():
                if _k.lower() in exclude_rules:
                    continue
                if 'untag'==_k:
                    continue
                if bool(_v):
                    continue
                filt_out.append(_k)
        return filt_out

    def _evaluate(self, rule):
        """
        递归执行list内容
        """
        operate_info = rule[0].split('#') # tag操作符号[eg: or#rule_1101]
        tag = operate_info[1] if len(operate_info)>1 else 'untag'
        operate = operate_info[0].lower()
        data = rule[1:]
        for id,idata in enumerate(data):
            if isinstance(idata, list):
                data[id] = self._evaluate(idata)
        iterate_out = self.get_func(operate)(*data)
        if 'untag'!=tag:
            self.whole_result.append({tag:iterate_out})
        return iterate_out

if __name__ == '__main__':
    engine = RuleParser()
    bb= engine.evaluate(_rule = ['in',3,2,5,4])
    print(bb)
    cc = engine.evaluate(_rule = ['in', 3, 2, 3, 4])
    print(cc)
    dd = engine.evaluate(_rule = ["and#result",[">#rule_10", 0, 0.05],["In#rule_20", 3, 2, 3, 4]])
    print(dd)


# coding=utf-8
# /usr/bin/env python

'''
Author: wenqiangw
Email: wenqiangw@opera.com
Date: 2020-04-16 16:50
Desc:
trans gps 到国家的城市地区,并画出geo图和热点图
'''
from .geo_map import GeoMap
from .google_map import plot_google_map


import pandas as pd
import numpy as np
import numbers
import pickle
import matplotlib.pyplot as plt
import seaborn as sns
from statsmodels.stats.outliers_influence import variance_inflation_factor
import warnings
warnings.filterwarnings("ignore")


def calc_woe(x,y,target_bad=1, eps=1e-4):
    bin = np.unique(x)
    prob_bad = np.array([np.logical_and(x==val,y==target_bad).sum() for val in bin])/float(np.sum(y==target_bad))
    prob_good = np.array([np.logical_and(x==val,y!=target_bad).sum() for val in bin])/float(np.sum(y!=target_bad))
    woe = np.log(np.maximum(prob_good,eps)/np.maximum(prob_bad,eps))
    iv = (prob_good-prob_bad)*woe
    return woe,iv

def CalcWOE(df, col, target):
    '''
    :param df: 包含需要计算WOE的变量和目标变量
    :param col: 需要计算WOE、IV的变量，必须是分箱后的变量，或者不需要分箱的类别型变量
    :param target: 目标变量，0、1表示好、坏
    :return: 返回WOE和IV
    '''
    total = df.groupby([col])[target].count()
    total = pd.DataFrame({'total': total})
    bad = df.groupby([col])[target].sum()
    bad = pd.DataFrame({'bad': bad})
    regroup = total.merge(bad, left_index=True, right_index=True, how='left')
    regroup.reset_index(level=0, inplace=True)
    N = sum(regroup['total'])
    B = sum(regroup['bad'])
    regroup['good'] = regroup['total'] - regroup['bad']
    G = N - B
    print(col,N,B)
    regroup['bad_pcnt'] = regroup['bad'].map(lambda x: x*1.0/B)
    regroup['good_pcnt'] = regroup['good'].map(lambda x: x * 1.0 / G)
    regroup['WOE'] = regroup.apply(lambda x: np.log(x.good_pcnt*1.0/x.bad_pcnt),axis = 1)
    WOE_dict = regroup[[col,'WOE']].set_index(col).to_dict(orient='index')
    for k, v in WOE_dict.items():
        WOE_dict[k] = v['WOE']
    IV = regroup.apply(lambda x: (x.good_pcnt-x.bad_pcnt)*np.log(x.good_pcnt*1.0/x.bad_pcnt),axis = 1)
    IV = sum(IV)
    return {"WOE": WOE_dict, 'IV':IV}



#读取数据
trainData = pd.read_csv('test1.csv')
##只计算统计变量
allFeatures = list(trainData.columns)
allFeatures.remove('Unnamed: 0')

##沿用之前分箱时的结果，详见之前
var_bin_list = []
less_value_features = []

WOE_dict = {}
IV_dict = {}
# 分箱后的变量进行编码，包括：
# 1，初始取值个数小于5，且不需要合并的类别型变量。存放在less_value_features中
# 2，初始取值个数小于5，需要合并的类别型变量。合并后新的变量存放在var_bin_list中
# 3，初始取值个数超过5，需要合并的类别型变量。合并后新的变量存放在var_bin_list中
# 4，连续变量。分箱后新的变量存放在var_bin_list中
all_var = var_bin_list + less_value_features
for var in all_var:
    woe_iv = CalcWOE(trainData, var, 'result')
    WOE_dict[var] = woe_iv['WOE']
    IV_dict[var] = woe_iv['IV']

file4 = open('WOE_dict.pkl', 'wb+')
pickle.dump(WOE_dict, file4)
file4.close()

# 将变量IV值进行降.,序排列，方便后续挑选变量
IV_dict_sorted = sorted(IV_dict.items(), key=lambda x: x[1], reverse=True)

IV_values = [i[1] for i in IV_dict_sorted]
IV_name = [i[0] for i in IV_dict_sorted]
plt.title('feature IV')
plt.bar(range(len(IV_values)), IV_values)
plt.show()

# 选取IV>0.02的变量
high_IV = {k: v for k, v in IV_dict.items() if v >= 0.02}
high_IV_sorted = sorted(high_IV.items(), key=lambda x: x[1], reverse=True)

short_list = high_IV.keys()
short_list_2 = []
for var in short_list:
    newVar = var + '_WOE'
    trainData[newVar] = trainData[var].map(WOE_dict[var])
    short_list_2.append(newVar)

# 对于上一步的结果，计算相关系数矩阵，并画出热力图进行数据可视化
trainDataWOE = trainData[short_list_2]
f, ax = plt.subplots(figsize=(10, 8))
corr = trainDataWOE.corr()
sns.heatmap(corr, mask=np.zeros_like(corr, dtype=np.bool), cmap=sns.diverging_palette(220, 10, as_cmap=True),
            square=True, ax=ax)

# 两两间的线性相关性检验
# 1，将候选变量按照IV进行降序排列
# 2，计算第i和第i+1的变量的线性相关系数
# 3，对于系数超过阈值的两个变量，剔除IV较低的一个
deleted_index = []
cnt_vars = len(high_IV_sorted)
for i in range(cnt_vars):
    if i in deleted_index:
        continue
    x1 = high_IV_sorted[i][0] + "_WOE"  # 'newVarWOe'
    for j in range(cnt_vars):
        if i == j or j in deleted_index:
            continue
        y1 = high_IV_sorted[j][0] + "_WOE"
        roh = np.corrcoef(trainData[x1], trainData[y1])[0, 1]
        if abs(roh) > 0.7:
            x1_IV = high_IV_sorted[i][1]
            y1_IV = high_IV_sorted[j][1]
            if x1_IV > y1_IV:
                deleted_index.append(j)
            else:
                deleted_index.append(i)

multi_analysis_vars_1 = [high_IV_sorted[i][0] + "_WOE" for i in range(cnt_vars) if i not in deleted_index]


## 多变量分析，一般不超过10合理，不合理的需要单独找出来剔除
X = np.matrix(trainData[multi_analysis_vars_1])
VIF_list = [variance_inflation_factor(X, i) for i in range(X.shape[1])]
max_VIF = max(VIF_list)
print(max_VIF)
multi_analysis = multi_analysis_vars_1


import pandas as pd
import datetime
import collections
import numpy as np
import numbers
import random
import sys
import seaborn as sns
import pickle
from itertools import combinations
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_curve
from sklearn.metrics import roc_auc_score
import statsmodels.api as sm
from importlib import reload
from matplotlib import pyplot as plt
from sklearn.linear_model import LogisticRegressionCV
from sklearn.linear_model import LogisticRegression
import warnings
warnings.filterwarnings("ignore")
from sklearn2pmml import sklearn2pmml,PMMLPipeline
from sklearn.ensemble import RandomForestClassifier
import tensorflow as tf
from tensorflow.contrib.learn.python.learn.estimators import SKCompat
from xgboost.sklearn import XGBClassifier
from sklearn.model_selection import GridSearchCV

def MissingCategorial(df,x):
    missing_vals = df[x].map(lambda x: int(x!=x))
    return sum(missing_vals)*1.0/df.shape[0]

def MissingContinuous(df,x):
    missing_vals = df[x].map(lambda x: int(np.isnan(x)))
    return sum(missing_vals) * 1.0 / df.shape[0]

def Outlier_Dectection(df,x):
    '''
    :param df:
    :param x:
    :return:
    '''
    p25, p75 = np.percentile(df[x], 25),np.percentile(df[x], 75)
    d = p75 - p25
    upper, lower =  p75 + 1.5*d, p25-1.5*d
    truncation = df[x].map(lambda x: max(min(upper, x), lower))
    return truncation


#读取数据
trainData = pd.read_csv('test1.csv')
##只计算统计变量
allFeatures = list(trainData.columns)
allFeatures.remove('Unnamed: 0')
allFeatures.remove('result')

#检查是否有常数型变量，并且检查是类别型还是数值型变量
numerical_var = []
for col in allFeatures:
    if len(set(trainData[col])) == 1:
        print('delete {} from the dataset because it is a constant'.format(col))
        del trainData[col]
        allFeatures.remove(col)
    else:
        uniq_valid_vals = [i for i in trainData[col] if i == i]
        uniq_valid_vals = list(set(uniq_valid_vals))
        if len(uniq_valid_vals) >= 10 and isinstance(uniq_valid_vals[0], numbers.Real):
            numerical_var.append(col)

categorical_var = [i for i in allFeatures if i not in numerical_var]

'''
对类别型变量，如果缺失超过80%, 就删除，否则保留。
'''
missing_pcnt_threshould_1 = 0.8
for col in categorical_var:
    missingRate = MissingCategorial(trainData,col)
    print('{0} has missing rate as {1}'.format(col,missingRate))
    if missingRate > missing_pcnt_threshould_1:
        categorical_var.remove(col)
        del trainData[col]
allData_bk = trainData.copy()

'''
用one-hot对类别型变量进行编码
'''
dummy_map = {}
dummy_columns = []
for raw_col in categorical_var:
    dummies = pd.get_dummies(trainData.loc[:, raw_col], prefix=raw_col)
    col_onehot = pd.concat([trainData[raw_col], dummies], axis=1)
    col_onehot = col_onehot.drop_duplicates()
    trainData = pd.concat([trainData, dummies], axis=1)
    del trainData[raw_col]
    dummy_map[raw_col] = col_onehot
    dummy_columns = dummy_columns + list(dummies)


'''
检查数值型变量
'''
missing_pcnt_threshould_2 = 0.8
deleted_var = []
for col in numerical_var:
    missingRate = MissingContinuous(trainData, col)
    print('{0} has missing rate as {1}'.format(col, missingRate))
    if missingRate > missing_pcnt_threshould_2:
        deleted_var.append(col)
        print('we delete variable {} because of its high missing rate'.format(col))
    else:
        if missingRate > 0:
            not_missing = trainData.loc[trainData[col] == trainData[col]][col]
            missing_position = trainData.loc[trainData[col] != trainData[col]][col].index
            not_missing_sample = random.sample(list(not_missing), len(missing_position))
            trainData.loc[missing_position,col] = not_missing_sample
            #del allData[col]
            #allData[col] = makeuped
            missingRate2 = MissingContinuous(trainData, col)
            print('missing rate after making up is:{}'.format(str(missingRate2)))

if deleted_var != []:
    for col in deleted_var:
        numerical_var.remove(col)
        del trainData[col]


'''
对极端值变量做处理。
'''
max_min_standardized = {}
for col in numerical_var:
    truncation = Outlier_Dectection(trainData, col)
    upper, lower = max(truncation), min(truncation)
    d = upper - lower
    if d == 0:
        print("{} is almost a constant".format(col))
        numerical_var.remove(col)
        continue
    trainData[col] = truncation.map(lambda x: (upper - x)/d)
    max_min_standardized[col] = [lower, upper]


all_features = list(trainData.columns)
all_features.remove('result')
X_train, y_train = trainData[all_features], trainData['result']

param_test1 = {'max_depth':range(3,10,2), 'min_child_weight':range(1,6,2)}
gsearch1 = GridSearchCV(estimator = XGBClassifier(learning_rate =0.1, n_estimators=100, gamma=0, subsample=0.8, colsample_bytree=0.8,
                                                  objective= 'binary:logistic',nthread=4,scale_pos_weight=1,seed=27),
                        param_grid = param_test1,scoring='roc_auc',n_jobs=4,iid=False,cv=5)
gsearch1.fit(X_train,y_train)
best_max_depth, best_min_child_weight = gsearch1.best_params_['max_depth'],gsearch1.best_params_['min_child_weight']

param_test2 = {'gamma':[i/10.0 for i in range(0,5)]}
gsearch2 = GridSearchCV(estimator = XGBClassifier(learning_rate =0.1, n_estimators=100, subsample=0.8, colsample_bytree=0.8, max_depth= best_max_depth,
                                                  min_child_weight=best_min_child_weight, objective= 'binary:logistic',nthread=4,scale_pos_weight=1,seed=27),
                        param_grid = param_test2,scoring='roc_auc',n_jobs=4,iid=False,cv=5)
gsearch2.fit(X_train,y_train)
best_gamma = gsearch2.best_params_['gamma']

param_test3 = {'subsample':[i/10.0 for i in range(6,10)],'colsample_bytree':[i/10.0 for i in range(6,10)]}
gsearch3 = GridSearchCV(estimator = XGBClassifier(learning_rate =0.1, n_estimators=100, max_depth= best_max_depth, gamma=best_gamma,
                                                  min_child_weight=best_min_child_weight, objective= 'binary:logistic',nthread=4,scale_pos_weight=1,seed=27),
                        param_grid = param_test3,scoring='roc_auc',n_jobs=4,iid=False,cv=5)
gsearch3.fit(X_train,y_train)
best_colsample_bytree, best_subsample = gsearch3.best_params_['colsample_bytree'], gsearch3.best_params_['subsample']

param_test4 = {'reg_alpha':[0.01,0.1,1,10,50,100,200,500]}
gsearch4 = GridSearchCV(estimator = XGBClassifier(learning_rate =0.1, n_estimators=100, max_depth= best_max_depth, gamma=best_gamma,
                                                  colsample_bytree = best_colsample_bytree, subsample = best_subsample,
                                                  min_child_weight=best_min_child_weight, objective= 'binary:logistic',nthread=4,scale_pos_weight=1,seed=27),
                        param_grid = param_test4,scoring='roc_auc',n_jobs=4,iid=False,cv=5)
gsearch4.fit(X_train,y_train)
best_reg_alpha = gsearch4.best_params_['reg_alpha']


param_test5 = {'n_estimators':range(100,401,10)}
gsearch5 = GridSearchCV(estimator = XGBClassifier(learning_rate =0.1,  max_depth= best_max_depth, gamma=best_gamma,
                                                  colsample_bytree = best_colsample_bytree, subsample = best_subsample,reg_alpha=best_reg_alpha,
                                                  min_child_weight=best_min_child_weight, objective= 'binary:logistic',nthread=4,scale_pos_weight=1,seed=27),
                        param_grid = param_test5,scoring='roc_auc',n_jobs=4,iid=False,cv=5)
gsearch5.fit(X_train,y_train)
best_n_estimators = gsearch5.best_params_

#用获取得到的最优参数再次训练模型
best_xgb = XGBClassifier(learning_rate =0.1, n_estimators=100, max_depth= best_max_depth, gamma=best_gamma,
                         colsample_bytree = best_colsample_bytree, subsample = best_subsample, reg_alpha=best_reg_alpha,
                         min_child_weight=best_min_child_weight, objective= 'binary:logistic',nthread=4,scale_pos_weight=1,seed=27)
best_xgb.fit(X_train,y_train)
y_pred = best_xgb.predict_proba(X_train)[:,1]
roc_auc_score(y_train, y_pred)
feature_importance = best_xgb.feature_importances_

#利用特征重要性筛去一部分无用的变量
X_train_temp = X_train.copy()
features_in_model = all_features
while(min(feature_importance)<0.00001):
    features_in_model = [features_in_model[i] for i in range(len(feature_importance)) if feature_importance[i] > 0.00001]
    X_train_temp= X_train_temp[features_in_model]
    best_xgb.fit(X_train_temp, y_train)
    feature_importance = best_xgb.feature_importances_

y_pred = best_xgb.predict_proba(X_train_temp)[:,1]
print(roc_auc_score(y_train, y_pred) ) 
print('There are {} features in the raw data'.format(X_train.shape[1]))
print('There are {} features in the reduced data'.format(X_train_temp.shape[1]))